@@ -967,7 +967,7 @@ def _ters_cutoff(dr, R, D) -> Array:
     cut-off values
   """
   outer = jnp.where(dr < R + D,
-                    0.5*(1.0 - jnp.sin(jnp.pi/2*(dr - R)/D)),
+                    0.5 * (1 - jnp.sin(jnp.pi / 2 * (dr - R) / D)),
                     0)
   inner = jnp.where(dr < R - D, 1, outer)
   return inner
@@ -1009,26 +1009,24 @@ def _ters_bij(R, D, c, d, h, lam3, beta, n, m,
   drij = space.distance(dRij)
   drik = space.distance(dRik)
 
+  mask_ijk *= (1 - jnp.eye(mask_ijk.shape[-1]))[None, :, :]
+
   # compute g_ijk - angle penalty value
   costheta = quantity.cosine_angles(dRij)
   gijk = 1.0 + (c**2 / d**2) - (c**2 / (d**2 + (h - costheta)**2))
-  gijk *= mask_ijk
-  gijk = vmap(lambda x: x.at[jnp.diag_indices(x.shape[0])].set(0))(gijk)
 
   # compute exponential term - distance penalty value
-  dr_diff = vmap(vmap(jnp.subtract, (None, 0)))(drij, drik)
+  dr_diff = drij[:, None, :] - drik[:, :, None]
+  dr_diff = jnp.where(mask_ijk, dr_diff, 0)
   explr3 = jnp.exp(lam3**m * dr_diff**m)
-  explr3 *= mask_ijk
-  explr3 = vmap(lambda x: x.at[jnp.diag_indices(x.shape[0])].set(0))(explr3)
 
   # compute fC with dr_ik
   fC = _ters_cutoff(drik, R, D)
 
   # compute zeta without diagonal term
-  tmp = jnp.multiply(gijk, explr3)
-  zeta_ij = jnp.einsum('ik,ikj->ij', fC, tmp)
-  bij = (1 + (beta*zeta_ij)**n)**(-1 / 2 / n)
-
+  prod = jnp.where(mask_ijk, gijk * explr3, 0)
+  zeta_ij = jnp.einsum('ik,ikj->ij', fC, prod)
+  bij = (1 + (beta * zeta_ij)**n)**(-1 / 2 / n)
   return bij
 
 def _ters_attractive(B: f64, lam2: f64,
@@ -1069,10 +1067,9 @@ def _ters_attractive(B: f64, lam2: f64,
 
   dr12 = space.distance(dR12)
   fC = _ters_cutoff(dr12, R, D)
-  fA = -B*jnp.exp(-lam2*dr12)
-  bij = _ters_bij(R, D, c, d, h, lam3, beta, n, m,
-                  dR12, dR13, mask_ijk)
-  return 0.5*fC*bij*fA
+  fA = -B * jnp.exp(-lam2 * dr12)
+  bij = _ters_bij(R, D, c, d, h, lam3, beta, n, m, dR12, dR13, mask_ijk)
+  return 0.5 * fC * bij * fA
 
 def _ters_repulsive(A: f64, lam1: f64, R: f64, D: f64,
                     dr: Array) -> Array:
@@ -1120,9 +1117,9 @@ def tersoff(displacement: DisplacementFn,
   if species is None:
     params = params[0]
   else:
-    raise NotImplementedError('Multiple species is not implemented yet.'
-                              ' Please raise an issue if this is important for'
-                              ' you.')
+    raise NotImplementedError('Multiple species is not implemented yet. '
+                              'Please raise an issue if this is important for '
+                              'you.')
 
   # define a repulsive and an attractive function with given parameters.
   repulsive_fn = partial(_ters_repulsive,
@@ -1134,29 +1131,36 @@ def tersoff(displacement: DisplacementFn,
                           params['cTf'], params['dTf'], params['hTf'],
                           params['lam3'], params['beta'],
                           params['nTf'], params['mTf'])
-
   # define compute functions.
   def compute_fn(R, **kwargs):
     d = partial(displacement, **kwargs)
-    dR = space.map_neighbor(d)(R, R)
+    dR = space.map_product(d)(R, R)
     dr = space.distance(dR)
     N = R.shape[0]
-    neighbor = jnp.broadcast_to(jnp.arange(N)[None, :], (N, N))
-    mask = partition.neighbor_list_mask(neighbor, mask_self=True)
+    mask = jnp.where(1 - jnp.eye(N), dr < params['R'] + params['D'], 0)
     mask_ijk = mask[:, None, :] * mask[:, :, None]
-    first_term = util.high_precision_sum(repulsive_fn(dr) * mask)
-    second_term = util.high_precision_sum(attractive_fn(dR, dR, mask_ijk)
-                                          * mask)
+    repulsive = util.safe_mask(mask, repulsive_fn, dr)
+    attractive = attractive_fn(dR, dR, mask_ijk) * mask
+    first_term = util.high_precision_sum(repulsive)
+    second_term = util.high_precision_sum(attractive)
     return first_term + second_term
   return compute_fn
 
-  return compute_fn
+
+def tersoff_from_lammps_parameters(
+    displacement: DisplacementFn,
+    f: TextIO,
+    ) -> Callable[[Array], Array]:
+  """Convenience wrapper to compute Tersoff energy with LAMMPS parameters."""
+  return tersoff(displacement, load_lammps_tersoff_parameters(f))
+
 
 def tersoff_neighbor_list(displacement: DisplacementFn,
                           box_size: float,
                           params: Array,
                           species: Optional[Array]=None,
                           dr_threshold: float=0.5,
+                          disable_cell_list: bool=False,
                           fractional_coordinates: bool=True,
                           format: NeighborListFormat=partition.Dense,
                           **neighbor_kwargs
@@ -1216,13 +1220,13 @@ def tersoff_neighbor_list(displacement: DisplacementFn,
       box_size,
       params['R'] + params['D'],
       dr_threshold,
-      disable_cell_list=True,
+      disable_cell_list=disable_cell_list,
       fractional_coordinates=fractional_coordinates,
       format=format,
       **neighbor_kwargs)
   else:
     raise NotImplementedError('Tersoff potential only implemented '
-                                'with Dense neighbor lists.')
+                              'with Dense neighbor lists.')
 
   # define compute functions
   def compute_fn(R, neighbor, **kwargs):
